import numpy as np
from numpy import pi, cos, sin


def thin_film_reflectance(index_film,
                          index_substrate,
                          film_thickness,
                          wavelength,
                          aoi=8.0,
                          polarization='mixed',
                          index_air=1.0003):
    """
    Calculate reflectance spectrum from a thin film. The optical stack
    consists of a semi-infinite layer of air, a thin film with a given
    thickness, and a semi-infinite substrate layer.

    Inputs can be float or ndarray. All inputs that are ndarray must have the
    same shape. For example, if an array of wavelength values with shape (1,
    N) is passed, then index_film can either be a float or an ndarray with
    shape (1,N).

    The method is derived from the matrix form of the Fresnel equations,
    detailed in Ref. [1].

    [1] https://arxiv.org/abs/1603.02720

    Parameters
    ----------

    index_film : ndarray or float

        Refractive index of the thin film at the values specified in the
        `wavelength` input.

    index_substrate : ndarray or float

        Refractive index of the substrate at the values specified in the
        `wavelength` input.

    film_thickness : ndarray or float

        Thickness of thin film in nm.

    wavelength : ndarray or float

        Wavelength in nm

    aoi : ndarray or float

        Angle of incidence of light onto thin film, in degrees.

    polarization : str

        Polarization of light, can be 's', 'p' or 'mixed'.

    index_air : ndarray or float

        Index of refraction of air, default is 1.0003

    Returns
    -------

    """

    # Calculate angles from Snell's law
    theta0 = aoi * pi / 180
    theta1 = np.arcsin(index_air / index_film * sin(theta0))
    theta2 = np.arcsin(index_air / index_substrate * sin(theta0))

    # Give index of refraction values better names.
    n0 = index_air
    n1 = index_film
    n2 = index_substrate

    # Calculate reflectance amplitudes
    rs01 = (n0 * cos(theta0) - n1 * cos(theta1)) / (
            n0 * cos(theta0) + n1 * cos(theta1))

    rp01 = (n1 * cos(theta0) - n0 * cos(theta1)) / (
            n1 * cos(theta0) + n0 * cos(theta1))

    rs12 = (n1 * cos(theta1) - n2 * cos(theta2)) / (
            n1 * cos(theta1) + n2 * cos(theta2))

    rp12 = (n2 * cos(theta1) - n1 * cos(theta2)) / (
            n2 * cos(theta1) + n1 * cos(theta2))

    # delta is used in film thickness.
    kz = 2 * pi * n1 * cos(theta1) / wavelength
    delta = film_thickness * kz

    # Calculate reflected power for s and p light.
    Rs = np.abs(
        np.exp(-1j * delta) * rs12 + np.exp(1j * delta) * rs01) ** 2 / np.abs(
        np.exp(-1j * delta) + np.exp(1j * delta) * rs01 * rs12) ** 2

    Rp = np.abs(
        np.exp(-1j * delta) * rp12 + np.exp(1j * delta) * rp01) ** 2 / np.abs(
        np.exp(-1j * delta) + np.exp(1j * delta) * rp01 * rp12) ** 2

    # Return reflectance depending on polarization chosen.
    if polarization == 'mixed':
        return 0.5 * (Rs + Rp)
    elif polarization == 's':
        return Rs
    elif polarization == 'p':
        return Rp
    else:
        raise Exception('Polarization must be "mixed", "s" or "p".')




def single_interface_reflectance(n0,n1, aoi=8.0, polarization='mixed'):
    """
    Reflection from a single interface for a wave traveling from material n0
    to n1. Both n0 and n1 can be an array of values.

    Parameters
    ----------
    n0 : ndarray or float
        Index of refraction for semi-infinite material with incident wave.

    n1 : ndarray or float
        Index of refraction for semi-infinite material that wave reflects from.

    aoi : ndarray or float
        Angle of incidence in degrees.

    polarization : str
        Polarization of wave, can be 's', 'p' or 'mixed'.

    Returns
    -------
    reflectance : ndarray
        Reflectance of interface.

    """
    theta0 = aoi * pi / 180
    theta1 = np.arcsin(n0 / n1 * sin(theta0))

    rs01 = (n0 * cos(theta0) - n1 * cos(theta1)) / (
            n0 * cos(theta0) + n1 * cos(theta1))
    rp01 = (n1 * cos(theta0) - n0 * cos(theta1)) / (
            n1 * cos(theta0) + n0 * cos(theta1))

    if polarization == 'mixed':
        return 0.5 * (np.abs(rs01) ** 2 + np.abs(rp01) ** 2)
    elif polarization == 's':
        return np.abs(rs01) ** 2
    elif polarization == 'p':
        return np.abs(rp01) ** 2
    else:
        raise Exception('Polarization must be "mixed", "s" or "p".')





import numpy as np
from scipy.optimize import minimize, basinhopping
from pvarc.materials import refractive_index_glass, refractive_index_porous_silica
from pvarc import single_interface_reflectance, thin_film_reflectance


def arc_reflection_model(wavelength,
                         thickness=125,
                         fraction_abraded=0,
                         porosity=0.3,
                         fraction_dust=0,
                         aoi=8,
                         n0=1.0003):
    """
    Return the reflection values for a model of an aged ARC. The reflection
    is a linear combination of reflection from a thin film of a variable
    thickness and the reflection from BK7 glass.

    Parameters
    ----------
    wavelength : ndarray

        wavelength in nm

    thickness

        thickness of ARC in nm.

    fraction_abraded

        fraction of coating loss. 1 corresponds to 100% of the coating area
        removed and only underlying glass is present, 0 corresponds to the
        coating covering the entire sample.

    fraction_dust

        fraction of module area covered by dust with reflectivity of 1. A
        value of 0 corresponds to no dust (clean sample), 1 to full dust
        coverage (reflectivity of 1).

    aoi

        angle of incidence in degrees.

    Returns
    -------

    reflectance : ndarray

        Reflectance of sample at the values of wavelength specified.

    """

    index_substrate = refractive_index_glass(wavelength)
    index_film = refractive_index_porous_silica(wavelength, porosity=porosity)
    glass_reflectance = single_interface_reflectance(
        n0=n0,
        n1=index_substrate,
        polarization='mixed',
        aoi=aoi)

    thin_film_R = thin_film_reflectance(index_film=index_film,
                                        index_substrate=index_substrate,
                                        film_thickness=thickness,
                                        aoi=aoi,
                                        wavelength=wavelength)

    reflectance = (1 - fraction_dust) * (
            fraction_abraded * glass_reflectance + (
            1 - fraction_abraded) * thin_film_R) + fraction_dust

    return reflectance


def estimate_arc_reflection_model_params(wavelength, reflectance, porosity=0.1,
                                         wavelength_min=450,
                                         wavelength_max=1000):
    # Smooth.
    # N = 5
    # reflectance = np.convolve(reflectance, np.ones((N,)), mode='same') / N

    cax = np.logical_and(wavelength > wavelength_min,
                         wavelength < wavelength_max)
    idx_min_reflection = np.argmin(reflectance[cax])

    reflection_min = reflectance[cax][idx_min_reflection]
    wavelength_min_reflection = wavelength[cax][idx_min_reflection]
    n = refractive_index_porous_silica(wavelength_min_reflection, porosity=0.5)
    thickness_estimate = 0.25 * wavelength_min_reflection / n

    # thickness_estimate = 120

    reflection_from_arc = float(
        arc_reflection_model(np.array([wavelength_min_reflection]),
                             thickness_estimate,
                             fraction_abraded=0,
                             fraction_dust=0,
                             porosity=porosity))
    reflection_from_glass = float(
        arc_reflection_model(np.array([wavelength_min_reflection]),
                             thickness_estimate,
                             fraction_abraded=1,
                             fraction_dust=0,
                             porosity=porosity))

    # print('Reflection from glass: {}'.format(reflection_from_glass))
    # print('Reflection from arc: {}'.format(reflection_from_arc))
    # print('Reflection from sample: {}'.format(reflection_min))

    fraction_abraded = (reflection_min - reflection_from_arc) / (
            reflection_from_glass - reflection_from_arc)

    # if thickness_estimate < 135:
    #     thickness_estimate = 135
    return {'thickness': thickness_estimate,
            'fraction_abraded': float(fraction_abraded),
            'fraction_dust': 0.001,
            'porosity': porosity
            }


def fit_arc_reflection_spectrum(wavelength,
                                reflectance,
                                x0=None,
                                aoi=8,
                                model='d',
                                fixed=None,
                                verbose=False,
                                method='basinhopping',
                                niter=20,
                                wavelength_min=450,
                                wavelength_max=1000):
    """
    This function fits an ARC model to the reflection spectrum. The ARC model
    is described in the function arc_reflection_model.

    Parameters
    ----------
    wavelength : ndarray

        Wavelength in nm

    reflectance : ndarray

        Fractional reflection between 0 and 1, unitless.

    x0 : dict

        Startpoint for fitting algorithm.

    aoi : float

        Angle of incidence of light

    model : str

        Different variation of the same model are available. The model is
        described under the function `arc_reflection_model`. The fixed values
        are specified in the input 'fixed'.

        'TP' - thickness and poristy are fit, fraction_abraded and fraction_dust
        set to fixed values.

        'TPA' - thickness, porosity and fraction_abraded are fit,
        fraction_dust is fixed.

        'TPAD' - thickness, porosity, fraction_abraded and fraction_dust are
        fit.

        'TAD' - thickness, fraction_abraded, fraction_dust are fit, porosity
        is a fixed value.

    fixed : dict

        Dictionary of parameters to be fixed. Default is:
           fixed = {'thickness': 125, 'fraction_abraded': 0, 'fraction_dust': 0,
             'porosity': 0.3}

     verbose : bool

        Whether to print output at each iteration.

    method : str

        Optimization method, can be 'minimize' or 'basinhopping'

    niter : int

        Number of basinhopping steps if method == 'basinhopping'

    wavelength_min : float

        Lower bound for wavelength values used in fit.

    wavelength_max : float

        Upper bound for wavelength values used in fit.

    Returns
    -------

    result : dict

        Dictionary of best fit values.

    ret

        Output of optimizer.

    """

    if np.mean(reflectance) > 1:
        print(
            'Warning: check that reflectance is a fractional value between 0 and 1.')

    if x0 == None:
        x0 = estimate_arc_reflection_model_params(wavelength, reflectance)

    fixed_default = {'thickness': 125,
                     'fraction_abraded': 0,
                     'fraction_dust': 0,
                     'porosity': 0.3}
    if fixed == None:
        fixed = fixed_default
    else:
        for p in fixed_default:
            if p not in fixed:
                fixed[p] = fixed_default[p]

    # print('x0: ', x0)

    scale = {'thickness': 0.01,
             'fraction_abraded': 1,
             'fraction_dust': 1000,
             'porosity': 1}

    if model == 'TPA':
        x0_list = [x0['thickness'] * scale['thickness'],
                   x0['fraction_abraded'] * scale['fraction_abraded'],
                   x0['porosity'] * scale['porosity']
                   ]
    elif model == 'TAD':
        x0_list = [x0['thickness'] * scale['thickness'],
                   x0['fraction_abraded'] * scale['fraction_abraded'],
                   x0['fraction_dust'] * scale['fraction_dust']
                   ]
    elif model == 'TPAD':
        x0_list = [x0['thickness'] * scale['thickness'],
                   x0['fraction_abraded'] * scale['fraction_abraded'],
                   x0['fraction_dust'] * scale['fraction_dust'],
                   x0['porosity'] * scale['porosity'],
                   ]
    elif model == 'TP':
        x0_list = [x0['thickness'] * scale['thickness'],
                   x0['porosity'] * scale['porosity'],
                   ]
    else:
        raise Exception('model options are "TP", "TPA", "TPAD" or "TAD"')

    # Increase by a factor of 100 to improve numeric accuracy.
    reflectance = reflectance * 100

    reflectance[reflectance < 0] = 0
    reflectance[reflectance > 100] = 100

    cax = np.logical_and(wavelength > wavelength_min,
                         wavelength < wavelength_max)

    wavelength_calc = wavelength.copy()

    index_substrate = refractive_index_glass(wavelength)

    glass_reflectance_calc = single_interface_reflectance(n0=1.0003,
                                                          n1=index_substrate,
                                                          aoi=aoi,
                                                          polarization='mixed',
                                                          )

    # # Get interpolator for correct
    # if not aoi == 8:
    #     raise Exception('aoi must be 8 degrees.')

    thickness_min = 50
    thickness_max = 200
    porosity_max = 0.499

    if model == 'TPA':
        bounds = [(thickness_min * scale['thickness'],
                   thickness_max * scale['thickness']),
                  (0, 1 * scale['fraction_abraded']),
                  (0, porosity_max * scale['porosity']),
                  ]
    elif model == 'TAD':
        bounds = [(thickness_min * scale['thickness'],
                   thickness_max * scale['thickness']),
                  (0, 1 * scale['fraction_abraded']),
                  (0, 1 * scale['fraction_dust']),
                  ]
    elif model == 'TPAD':
        bounds = [(thickness_min * scale['thickness'],
                   thickness_max * scale['thickness']),
                  (0, 1 * scale['fraction_abraded']),
                  (0, 1 * scale['fraction_dust']),
                  (0, porosity_max * scale['porosity']),
                  ]
    elif model == 'TP':
        bounds = [(thickness_min * scale['thickness'],
                   thickness_max * scale['thickness']),
                  (0, porosity_max * scale['porosity']),
                  ]

    def arc_model_c(wavelength, thickness, fraction_abraded, fraction_dust,
                    porosity):
        index_film = refractive_index_porous_silica(wavelength=wavelength,
                                         porosity=porosity)

        index_substrate = refractive_index_glass(wavelength=wavelength)
        thin_film_R = thin_film_reflectance(
            index_film=index_film,
            index_substrate=index_substrate,
            film_thickness=thickness,
            wavelength=wavelength,
            aoi=aoi
        )
        #
        # thin_film_R = thin_film_reflection_fast(
        #     wavelength=wavelength,
        #     thickness=thickness,
        #     aoi=aoi,
        #     porosity=porosity)

        #
        # index_film = index_porous_silica(wavelength=wavelength,
        #                                  porosity=porosity)
        # thin_film_reflectance = thin_film_reflection(
        #     polarization='mixed',
        #     wavelength=wavelength,
        #     d_list=[np.inf, thickness, np.inf],
        #     index_film=index_film,
        #     index_substrate=index_glass,
        #     aoi=aoi)

        glass_reflectance = np.interp(wavelength, wavelength_calc,
                                      glass_reflectance_calc)

        reflectance = (1 - fraction_dust) * (
                fraction_abraded * glass_reflectance + (
                1 - fraction_abraded) * thin_film_R) + fraction_dust

        return 100 * reflectance

    def arc_coating_error_function(x):

        # print('x: ',x)

        if model == 'TPA':
            thickness = x[0] / scale['thickness']
            fraction_abraded = x[1] / scale['fraction_abraded']
            porosity = x[2] / scale['porosity']
            reflectance_model = arc_model_c(wavelength, thickness,
                                            fraction_abraded=fraction_abraded,
                                            fraction_dust=fixed[
                                                'fraction_dust'],
                                            porosity=porosity)
        elif model == 'TAD':
            thickness = x[0] / scale['thickness']
            fraction_abraded = x[1] / scale['fraction_abraded']
            fraction_dust = x[2] / scale['fraction_dust']
            reflectance_model = arc_model_c(wavelength, thickness,
                                            fraction_abraded, fraction_dust,
                                            porosity=fixed['porosity'])
            # if verbose:
            #     print(
            #         'Thickness: {:03.2f}, Fraction Abraded: {:.1%}, Fraction dust: {:.1%}'.format(
            #             thickness,
            #             fraction_abraded,
            #             fraction_dust))
        elif model == 'TPAD':

            thickness = x[0] / scale['thickness']
            fraction_abraded = x[1] / scale['fraction_abraded']
            fraction_dust = x[2] / scale['fraction_dust']
            porosity = x[3] / scale['porosity']
            reflectance_model = arc_model_c(wavelength, thickness,
                                            fraction_abraded, fraction_dust,
                                            porosity)
        elif model == 'TP':

            thickness = x[0] / scale['thickness']
            porosity = x[1] / scale['porosity']
            reflectance_model = arc_model_c(wavelength, thickness,
                                            fraction_abraded=fixed[
                                                'fraction_abraded'],
                                            fraction_dust=fixed[
                                                'fraction_dust'],
                                            porosity=porosity)

        else:
            raise Exception('model type unknown')

        residual = np.mean(
            np.sqrt(np.abs(reflectance_model - reflectance) ** 2))

        return residual

    if method == 'minimize':
        res = minimize(arc_coating_error_function,
                       x0=x0_list,
                       options=dict(
                           # maxiter=100,
                           disp=verbose
                       ),
                       bounds=bounds
                       )
    elif method == 'basinhopping':

        def basinhopping_callback(x, f, accept):
            if model == 'TPA' and verbose:
                # print('x:', x)
                print(
                    '--\nThickness: {:03.2f}, Fraction Abraded: {:.1%}, Porosity: {:.1%}'.format(
                        x[0] / scale['thickness'],
                        x[1] / scale['fraction_abraded'],
                        x[2] / scale['porosity']
                    )
                )

        res = basinhopping(arc_coating_error_function,
                           x0=x0_list,
                           niter=niter,
                           minimizer_kwargs={'bounds': bounds},
                           disp=verbose,
                           callback=basinhopping_callback
                           )

    if model == 'TPA':
        result = {'thickness': res['x'][0] / scale['thickness'],
                  'fraction_abraded': res['x'][1] / scale['fraction_abraded'],
                  'fraction_dust': fixed['fraction_dust'],
                  'porosity': res['x'][2] / scale['porosity']}
    elif model == 'TAD':
        result = {'thickness': res['x'][0] / scale['thickness'],
                  'fraction_abraded': res['x'][1] / scale['fraction_abraded'],
                  'fraction_dust': res['x'][2] / scale['fraction_dust'],
                  'porosity': fixed['porosity']}
    elif model == 'TPAD':
        result = {'thickness': res['x'][0] / scale['thickness'],
                  'fraction_abraded': res['x'][1] / scale['fraction_abraded'],
                  'fraction_dust': res['x'][2] / scale['fraction_dust'],
                  'porosity': res['x'][3] / scale['porosity'],
                  }
    elif model == 'TP':
        result = {'thickness': res['x'][0] / scale['thickness'],
                  'fraction_abraded': fixed['fraction_abraded'],
                  'fraction_dust': fixed['fraction_dust'],
                  'porosity': res['x'][1] / scale['porosity'],
                  }

    return result, res




import numpy as np
import pandas as pd
import os

dir_path = os.path.dirname(os.path.realpath(__file__))


def refractive_index_imaginary_silica(wavelength):
    dir_path = os.path.dirname(os.path.realpath(__file__))
    fname = os.path.join(dir_path, 'materials', 'refractive_index_k_silica_Kitamura2007.csv')
    df = pd.read_csv(fname)
    print(df.keys())
    k = np.interp(wavelength, df['wavelength'] * 1000., df['k'])

    return k

def reflectance_PTFE(wavelength):


    dir_path = os.path.dirname(os.path.realpath(__file__))
    fname = os.path.join(dir_path, 'materials', 'PMR10_PTFE_reflectance.xlsx')
    df = pd.read_excel(fname)
    reflectance =  np.interp(wavelength, df['Wavelength (nm)'], df['Reflectance (%)'])

    return reflectance


def refractive_index_fused_silica(wavelength):
    """
    Refractive index for Corning UV fused silica 7980

    https://www.corning.com/media/worldwide/csm/documents/5bf092438c5546dfa9b08e423348317b.pdf

    Parameters
    ----------
    wavelength
        wavelength in nm

    Returns
    -------

    """
    wavelength_um = wavelength / 1000

    A0 = 2.104025406E+00
    A1 = -1.456000330E-04
    A2 = -9.049135390E-03
    A3 = 8.801830992E-03
    A4 = 8.435237228E-05
    A5 = 1.681656789E-06
    A6 = -1.675425449E-08
    A7 = 8.326602461E-10

    n = np.sqrt( A0 + A1 * wavelength_um ** 4 + A2 * wavelength_um ** 2 + A3 * wavelength_um ** -2 + \
         A4 * wavelength_um ** -4 + A5 * wavelength_um ** -6 + A6 * wavelength_um ** -8 + A7 * wavelength_um ** -10 )

    return n


def refractive_index_glass(wavelength, type='soda-lime-low-iron'):
    """
    Return real part of refractive index for glass given an array of
    wavelengths.

    Data for Soda-lime glass from:
     https://refractiveindex.info/?shelf=glass&book=soda-lime&page=Rubin-clear

    Rubin 1985. Range of validity is 310-4600 nm.

    Data for BK7 glass from:
        https://refractiveindex.info/?shelf=glass&book=BK7&page=SCHOTT

    Parameters
    ----------
    wavelength : ndarray
        wavelength in nm

    type : str
        Type of glass. Options are:

        'soda-lime-low-iron'

    Returns
    -------
    index_of_refraction : ndarray
        Index of refraction at specified wavelength.

    """

    if type.lower() == 'soda-lime-low-iron':
        wavelength = wavelength / 1000
        n = 1.5130 - 0.003169 * wavelength ** 2 + 0.003962 * wavelength ** -2 + 0 * 1j

        # n[wavelength < 0.3] = n[wavelength < 0.3] + 1j*0
    elif type.upper() == 'BK7':
        wavelength = wavelength / 1000
        n = np.sqrt(1 + \
                    (1.03961212 * wavelength ** 2) / (
                            wavelength ** 2 - 0.00600069867) + \
                    (0.231792344 * wavelength ** 2) / (
                            wavelength ** 2 - 0.0200179144) + \
                    (1.01046945 * wavelength ** 2) / (
                            wavelength ** 2 - 103.560653)
                    )

    return n


def refractive_index_porous_silica(wavelength, porosity=0.5):
    """
    Calculates index of refraction for porous silica using the effective
    medium approximation and volume averaging theory.

    https://refractiveindex.info/?shelf=main&book=SiO2&page=Malitson

    Parameters
    ----------
    wavelength : numeric

        Wavelength in nm.

    porosity : float

        Fractional porosity, a number from 0 to 1.0.


    Returns
    -------

    index : numeric

        Refractive index

    """
    wavelength_um = wavelength / 1000
    n = np.sqrt(1 + \
                (0.6961663 * wavelength_um ** 2) / (
                        wavelength_um ** 2 - 0.06840432 ** 2) + \
                (0.4079426 * wavelength_um ** 2) / (
                        wavelength_um ** 2 - 0.11624142 ** 2) + \
                (0.8974794 * wavelength_um ** 2) / (
                        wavelength_um ** 2 - 9.8961612 ** 2)
                )
    n_air = 1.00029

    n_total = np.sqrt(n ** 2 * (1 - porosity) + n_air ** 2 * (porosity)) + 0 * 1j

    # k0 = 5e-6
    # k1 = 5e-7
    # wavelength0 = 0.31
    # wavelength1 = 0.36

    # n_total = n_total + 1j*refractive_index_imaginary_silica(wavelength)*1e4
    # n_total = n_total + 1j*np.exp( np.log(k0) + np.log(k1) * (wavelength - wavelength0)/(wavelength1-wavelength0))

    return n_total



import numpy as np
import pandas as pd

import colour
from colour.colorimetry import (MultiSpectralDistributions, SpectralShape, sd_ones)

from colour import SpectralDistribution, CubicSplineInterpolator, Extrapolator, \
    XYZ_to_sRGB, sd_to_XYZ, XYZ_to_xyY, XYZ_to_CIECAM02, xyY_to_XYZ, SDS_ILLUMINANTS



from colour.utilities.array import is_uniform, interval

from pvarc.materials import refractive_index_porous_silica, refractive_index_glass
from pvarc import thin_film_reflectance


# from skimage.color import rgb2xyz, xyz2lab


from colour.colorimetry import (intermediate_lightness_function_CIE1976,
                                intermediate_luminance_function_CIE1976)
from colour.models import xy_to_xyY, xyY_to_XYZ, Jab_to_JCh, JCh_to_Jab
from colour.utilities import (from_range_1, from_range_100, to_domain_1,
                              to_domain_100, tsplit, tstack)


def XYZ_to_Lab(
        XYZ,
        XYZ_n,
        ):
    """
    Converts from *CIE XYZ* tristimulus values to *CIE L\\*a\\*b\\**
    colourspace.

    ** TODD KARIN: Modified to use XYZ_n rather than illuminant.

    Parameters
    ----------
    XYZ : array_like
        *CIE XYZ* tristimulus values.
    illuminant : array_like, optional
        Reference *illuminant* *CIE xy* chromaticity coordinates or *CIE xyY*
        colourspace array.

    Returns
    -------
    ndarray
        *CIE L\\*a\\*b\\** colourspace array.

    Notes
    -----

    +----------------+-----------------------+-----------------+
    | **Domain**     | **Scale - Reference** | **Scale - 1**   |
    +================+=======================+=================+
    | ``XYZ``        | [0, 1]                | [0, 1]          |
    +----------------+-----------------------+-----------------+
    | ``illuminant`` | [0, 1]                | [0, 1]          |
    +----------------+-----------------------+-----------------+

    +----------------+-----------------------+-----------------+
    | **Range**      | **Scale - Reference** | **Scale - 1**   |
    +================+=======================+=================+
    | ``Lab``        | ``L`` : [0, 100]      | ``L`` : [0, 1]  |
    |                |                       |                 |
    |                | ``a`` : [-100, 100]   | ``a`` : [-1, 1] |
    |                |                       |                 |
    |                | ``b`` : [-100, 100]   | ``b`` : [-1, 1] |
    +----------------+-----------------------+-----------------+

    References
    ----------
    :cite:`CIETC1-482004m`

    Examples
    --------
    >>> import numpy as np
    >>> XYZ = np.array([0.20654008, 0.12197225, 0.05136952])
    >>> XYZ_to_Lab(XYZ)  # doctest: +ELLIPSIS
    array([ 41.5278752...,  52.6385830...,  26.9231792...])
    """

    X, Y, Z = tsplit(to_domain_1(XYZ))

    X_n, Y_n, Z_n = tsplit(XYZ_n)

    f_X_X_n = intermediate_lightness_function_CIE1976(X, X_n)
    f_Y_Y_n = intermediate_lightness_function_CIE1976(Y, Y_n)
    f_Z_Z_n = intermediate_lightness_function_CIE1976(Z, Z_n)

    L = 116 * f_Y_Y_n - 16
    a = 500 * (f_X_X_n - f_Y_Y_n)
    b = 200 * (f_Y_Y_n - f_Z_Z_n)

    Lab = tstack([L, a, b])

    return from_range_100(Lab)

def spectrum_to_XYZ(wavelength, spectrum,
                    # cmfs=None,
                    illuminant=None,
                    ):
    """
    Calculate the rgb color given a wavelength and spectrum

    Parameters
    ----------
    wavelength
    spectrum
    cmfs
    illuminant

    Returns
    -------

    """
    # if cmfs is None:
        # cmfs = colour.STANDARD_OBSERVERS_CMFS['CIE 1931 2 Degree Standard Observer']



    if illuminant is None:
        # illuminant = ILLUMINANTS_SDS['CIE 1931 2 Degree Standard Observer']['D65']
        illuminant = sd_ones()
    elif type(illuminant)==type(''):
        illuminant = SDS_ILLUMINANTS[illuminant]
    # # Get illuminant
    # if type(illuminant) == type(''):
    #     illuminant = colour.ILLUMINANTS_SDS[illuminant]

    # Build spectral distribution object
    sd = SpectralDistribution(pd.Series(spectrum, index=wavelength),
                              interpolator=CubicSplineInterpolator,
                              extrapolator=Extrapolator)

    # Calculate xyz color coordinates.
    xyz = sd_to_XYZ(sd=sd,
                    illuminant=illuminant)

    return xyz


def spectrum_to_rgb(wavelength, spectrum,
                    # cmfs=None,
                    illuminant='LED-B3',
                    Y=None,
                    ):
    """
    Calculate the rgb color given a wavelength and spectrum. Note spectrum should have spacing of 1, 5, 10 or 10 nm.

    Parameters
    ----------
    wavelength
    spectrum
    cmfs
    illuminant

    Returns
    -------

    """

    # if is_uniform(wavelength) and interval(wavelength)
    XYZ = spectrum_to_XYZ(wavelength=wavelength,
                          spectrum=spectrum,
                          # cmfs=cmfs,
                          illuminant=illuminant)

    XYZ = XYZ / 100

    if Y is not None:
        xyY = XYZ_to_xyY(XYZ)
        xyY[2] = Y
        XYZ = xyY_to_XYZ(xyY)

    # Convert to rgb. Note the illuminant doesn't matter here except for points
    # where XYZ=0.
    rgb = XYZ_to_sRGB(XYZ)

    return rgb


def spectrum_to_xyY(wavelength, spectrum,
                    cmfs=None,
                    illuminant='LED-B3',
                    ):
    """
    Calculate the rgb color given a wavelength and spectrum

    Parameters
    ----------
    wavelength
    spectrum
    cmfs
    illuminant

    Returns
    -------

    """
    xyz = spectrum_to_XYZ(wavelength=wavelength,
                          spectrum=spectrum,
                          # cmfs=cmfs,
                          illuminant=illuminant)

    # Convert to rgb.
    # TODO: check on the divide by 100
    xyY = XYZ_to_xyY(xyz / 100)

    return xyY


# def spectrum_to_Lab(wavelength, spectrum,
#                     cmfs=None,
#                     illuminant='LED-B3',
#                     ):
#     xyz = spectrum_to_XYZ(wavelength=wavelength,
#                           spectrum=spectrum,
#                           cmfs=cmfs,
#                           illuminant=illuminant)
#
#     # Convert to rgb.
#     # TODO: check on the divide by 100
#     Lab = XYZ_to_Lab(xyz / 100)
#
#     return Lab


def arc_model_to_rgb(thickness, porosity, aoi=8):
    # Wavelength axis
    wavelength = np.arange(360, 801, 5).astype('float')

    # Choice of illuminant makes a very small change to the calculated RGB color.
    illuminant = 'LED-B3'

    # # Scan thickness and porosity.
    # thickness = np.arange(0, 196, 5).astype('float')
    # porosity = np.arange(0, 0.51, 0.1).astype('float')

    # col = np.zeros((3, len(thickness), len(porosity)))
    # col_hex = np.empty((len(thickness), len(porosity)), dtype='object')
    # xyY = np.zeros((3, len(thickness), len(porosity)))

    index_film = refractive_index_porous_silica(wavelength, porosity)
    index_substrate = refractive_index_glass(wavelength)

    # Calculate reflectance
    reflectance = thin_film_reflectance(index_film=index_film,
                                        index_substrate=index_substrate,
                                        film_thickness=thickness,
                                        aoi=aoi,
                                        wavelength=wavelength)
    reflectance_ref = thin_film_reflectance(index_film=index_film,
                                            index_substrate=index_substrate,
                                            film_thickness=0,
                                            aoi=aoi,
                                            wavelength=wavelength)

    rgb = spectrum_to_rgb(wavelength, reflectance,
                          illuminant=illuminant)
    rgb_ref = spectrum_to_rgb(wavelength, reflectance_ref,
                              illuminant=illuminant)

    # White balance
    rgb_wb = rgb / rgb_ref

    # Clamp
    rgb_wb[rgb_wb >= 1] = 1
    rgb_wb[rgb_wb < 0] = 0

    return rgb_wb

#
# def rgb2xyY(rgb):
#     xyz = rgb2xyz(rgb)



import numpy as np
import pandas as pd
import os
# import matplotlib

# matplotlib.use('TkAgg')
# import matplotlib.pyplot as plt
from tqdm import tqdm
from pvarc import thin_film_reflectance, single_interface_reflectance
from pvarc.materials import refractive_index_porous_silica, \
    refractive_index_glass
# from pvarc.color import spectrum_to_rgb, spectrum_to_XYZ, spectrum_to_xyY
from pvarc.metrics import solar_weighted_photon_reflectance
# from glob import glob
#
# from pvarc.oceaninsight import read_oceanview_file
#
# from colour import SpectralDistribution, CubicSplineInterpolator, Extrapolator, \
#     XYZ_to_sRGB, sd_to_XYZ, XYZ_to_xyY, XYZ_to_CIECAM02, xyY_to_XYZ, xy_to_xyY, \
#     RGB_to_XYZ, XYZ_to_xyY, \
#     Lab_to_XYZ, ILLUMINANTS, ILLUMINANTS_SDS, XYZ_to_RGB, sRGB_to_XYZ
#
# from colour import xy_to_xyY, RGB_to_XYZ, XYZ_to_xyY
# from skimage.color import rgb2xyz
from scipy.interpolate import griddata, RegularGridInterpolator


def calculate_rgb_vs_thickness_porosity(
        camera='Ximea-MC050cg_combined_labeled.csv',
        light_source='LEDW7E_A01_intensity.csv',
        optical_system='MVL23M23.csv',
        thickness_max=186,
        thickness_step=0.4,
        porosity_max=0.5,
        porosity_step=0.01,
        aoi=0):
    # Wavelength axis
    wavelength = np.arange(300, 755, 5).astype('float')
    dwavelength = wavelength[1] - wavelength[0]

    # Scan thickness and porosity.
    thickness = np.arange(0, thickness_max, thickness_step).astype('float')
    porosity = np.arange(0, porosity_max, porosity_step).astype('float')

    # Initialize arrays.
    swpr = np.zeros((len(thickness), len(porosity)))
    rgb_wb = np.zeros((len(thickness), len(porosity), 3))

    # Load Camera QE
    qe_fpath = os.path.join(os.path.dirname(__file__), 'cameras',
                            camera)
    df = pd.read_csv(qe_fpath, skiprows=2)
    dfi = pd.DataFrame({'Wavelength': wavelength})
    for k in ['Red', 'Green', 'Blue']:
        dfi[k] = np.interp(wavelength, df['Wavelength'], df[k],
                           left=0, right=0)

    # Load Illuminant spectrum
    illum_fpath = os.path.join(os.path.dirname(__file__), 'sources',
                               light_source)
    df_illum = pd.read_csv(illum_fpath, skiprows=2)
    # illuminant_sd = ILLUMINANTS_SDS[sources[s]]
    # illuminant_conv = ILLUMINANTS['CIE 1931 2 Degree Standard Observer'][illuminant_name]

    illuminant_spectrum = np.interp(wavelength, df_illum['Wavelength'],
                                    df_illum['Intensity'], left=0, right=0)

    illuminant_spectrum_photon = illuminant_spectrum * wavelength

    # Load optical system
    optical_system_fpath = os.path.join(os.path.dirname(__file__), 'cameras',
                                        optical_system)
    df_optical_system = pd.read_csv(optical_system_fpath, skiprows=2)
    optical_system_transmission = 0.01 * np.interp(
        wavelength,
        df_optical_system['Wavelength'],
        df_optical_system['Transmission'],
        left=0, right=0)

    # Calculate RGB colors
    for k in tqdm(range(len(porosity))):

        index_film = refractive_index_porous_silica(wavelength, porosity[k])
        index_substrate = refractive_index_glass(wavelength)

        for j in range(len(thickness)):

            # Calculate reflectance
            reflectance = thin_film_reflectance(index_film=index_film,
                                                index_substrate=index_substrate,
                                                film_thickness=thickness[j],
                                                aoi=aoi,
                                                wavelength=wavelength)
            # for c in ['Blue', 'Green', 'Red']:
            #     np.sum(reflectance * illuminant_spectrum_photon * dfi[c] / 100) * dwavelength

            rgb = np.sum(
                reflectance[:, np.newaxis] * \
                optical_system_transmission[:, np.newaxis] * \
                illuminant_spectrum_photon[:, np.newaxis] * \
                np.array(dfi.iloc[:, 1:]) / 100,
                axis=0) * dwavelength

            swpr[j, k] = solar_weighted_photon_reflectance(wavelength,
                                                           reflectance)

            # Use first run through, with 0 nm thickness, (i.e. low-iron glass)
            # as a reference for white balance.
            if thickness[j] == 0:
                rgb_ref = rgb.copy()

            # White balance
            rgb_wb[j, k, :] = rgb / rgb_ref

    # x = rgb_wb[0, :, :] / np.sum(rgb_wb, axis=0)
    # y = rgb_wb[1, :, :] / np.sum(rgb_wb, axis=0)
    # z = rgb_wb[2, :, :] / np.sum(rgb_wb, axis=0)

    # t_grid, P_grid = np.meshgrid(thickness, porosity, indexing='ij')

    return thickness, porosity, rgb_wb, swpr


def build_rgb_to_thickness_porosity_interpolator_data(
        camera='Ximea-MC050cg_combined_labeled.csv',
        light_source='LEDW7E_A01_intensity.csv',
        thickness_max=500,
        thickness_step=1,
        porosity_max=0.5,
        porosity_step=0.01):
    thickness, porosity, rgb_wb, swpr = calculate_rgb_vs_thickness_porosity(
        camera=camera,
        light_source=light_source,
        thickness_max=thickness_max,
        thickness_step=thickness_step,
        porosity_max=porosity_max,
        porosity_step=porosity_step)

    x = rgb_wb[:, :, 0] / np.sum(rgb_wb, axis=2)
    y = rgb_wb[:, :, 1] / np.sum(rgb_wb, axis=2)
    t_grid, P_grid = np.meshgrid(thickness, porosity, indexing='ij')

    # Save data to file:
    df = pd.DataFrame(dict(
        x=x.flatten(),
        y=y.flatten(),
        R=rgb_wb[:, :, 0].flatten(),
        G=rgb_wb[:, :, 1].flatten(),
        B=rgb_wb[:, :, 2].flatten(),
        thickness=t_grid.flatten(),
        porosity=P_grid.flatten(),
        swpr=swpr.flatten(),
    ))

    df.to_csv(get_thickness_porosity_interpolator_filename(),
              index=False)


def get_thickness_porosity_interpolator_filename():
    return os.path.join(os.path.dirname(__file__),
                        'rgb_to_thickness_porosity_data.csv')


def get_thickness_porosity_interpolator_data():
    """
    Get data on how coating color (RGB) depends on thickness and porosity.

    Returns
    -------
    dataframe with columns

        'thickness': coating thickness in nm

        'R': Red response, white balanced to a zero thickness coating.

        'G': Green response, white balanced to a zero thickness coating.

        'B': Blue response, white balanced to a zero thickness coating.

        'x': R/(R+G+B)

        'y': G/(R+G+B)

        'swpr': Solar-weighted photon reflectance.
    """
    return pd.read_csv(get_thickness_porosity_interpolator_filename())


# Not the best way to do this, but it avoids loading data and building the
# interpolator every time the following functions are called.
if os.path.exists(get_thickness_porosity_interpolator_filename()):
    _tp_rgb = get_thickness_porosity_interpolator_data()

    _thickness_list = np.unique(_tp_rgb['thickness'])
    _porosity_list = np.unique(_tp_rgb['porosity'])

    _tp_shape = (len(_thickness_list), len(_porosity_list))
    _thickness_grid = np.array(_tp_rgb['thickness']).reshape(_tp_shape)
    _porosity_grid = np.array(_tp_rgb['porosity']).reshape(_tp_shape)
    _R_grid = np.array(_tp_rgb['R']).reshape(_tp_shape)
    _G_grid = np.array(_tp_rgb['G']).reshape(_tp_shape)
    _B_grid = np.array(_tp_rgb['B']).reshape(_tp_shape)

    _R_interpolator = RegularGridInterpolator((_thickness_list, _porosity_list),
                                              _R_grid,
                                              bounds_error=False,
                                              fill_value=np.nan)
    _G_interpolator = RegularGridInterpolator((_thickness_list, _porosity_list),
                                              _G_grid, bounds_error=False,
                                              fill_value=np.nan)
    _B_interpolator = RegularGridInterpolator((_thickness_list, _porosity_list),
                                              _B_grid, bounds_error=False,
                                              fill_value=np.nan)
else:
    print('Building thickness porosity interpolator data...')
    build_rgb_to_thickness_porosity_interpolator_data()


def calculate_rgb(thickness, porosity):
    coords = np.concatenate((np.atleast_1d(thickness).flatten()[:, np.newaxis],
                             np.atleast_1d(porosity).flatten()[:, np.newaxis]),
                            axis=1)

    rgb = np.concatenate((
        _R_interpolator(coords)[:, np.newaxis],
        _G_interpolator(coords)[:, np.newaxis],
        _B_interpolator(coords)[:, np.newaxis]
    ), axis=1)

    return rgb


def calculate_thickness(x, y, xy_distance_white=0.03,
                        porosity_min=0,
                        porosity_max=0.5,
                        ):
    """

    Parameters
    ----------
    x
    y
    xy_distance_white

    Returns
    -------

    """

    df = get_thickness_porosity_interpolator_data()
    df = df[np.logical_and(df['porosity'] > porosity_min,
                           df['porosity'] < porosity_max)]
    xa = np.atleast_1d(x)
    ya = np.atleast_1d(y)

    thickness_flat = griddata(np.array([df['x'], df['y']]).transpose(),
                              df['thickness'],
                              np.array(
                                  [xa.flatten(), ya.flatten()]).transpose())

    white_idx = (xa.flatten() - 1 / 3) ** 2 + (
            ya.flatten() - 1 / 3) ** 2 < xy_distance_white ** 2
    thickness_flat[white_idx] = 0

    thickness = np.reshape(thickness_flat, xa.shape)

    if np.isscalar(x) and np.isscalar(y):
        thickness = thickness[0]

    return thickness


def calculate_porosity(x, y,
                       xy_distance_white=0.03,
                       porosity_min=0,
                       porosity_max=0.5,
                       ):
    df = get_thickness_porosity_interpolator_data()
    df = df[np.logical_and(df['porosity'] > porosity_min,
                           df['porosity'] < porosity_max)]

    xa = np.atleast_1d(x)
    ya = np.atleast_1d(y)

    porosity_flat = griddata(np.array([df['x'], df['y']]).transpose(),
                             df['porosity'],
                             np.array([xa.flatten(), ya.flatten()]).transpose())

    white_idx = (xa.flatten() - 1 / 3) ** 2 + (
            ya.flatten() - 1 / 3) ** 2 < xy_distance_white ** 2
    porosity_flat[white_idx] = 0

    porosity = np.reshape(porosity_flat, xa.shape)

    if np.isscalar(x) and np.isscalar(y):
        porosity = porosity[0]

    return porosity


def calculate_swpr(x, y, xy_distance_white=0.03,
                   porosity_min=0,
                   porosity_max=0.5,
                   ):
    df = get_thickness_porosity_interpolator_data()
    df = df[np.logical_and(df['porosity'] > porosity_min,
                           df['porosity'] < porosity_max)]

    xa = np.atleast_1d(x)
    ya = np.atleast_1d(y)

    swpr_flat = griddata(np.array([df['x'], df['y']]).transpose(),
                         df['swpr'],
                         np.array([xa.flatten(), ya.flatten()]).transpose())

    white_idx = (xa.flatten() - 1 / 3) ** 2 + (
            ya.flatten() - 1 / 3) ** 2 < xy_distance_white ** 2
    swpr_flat[white_idx] = df.iloc[0]['swpr']

    swpr = np.reshape(swpr_flat, xa.shape)

    if np.isscalar(x) and np.isscalar(y):
        swpr = swpr[0]

    return swpr
